@@ -3,12 +3,14 @@ package XML::Sig;
 use strict;
 use warnings;
 
+use Encode;
+# ABSTRACT: XML::Sig - A toolkit to help sign and verify XML Digital Signatures
 
 # use 'our' on v5.6.0
 use vars qw($VERSION @EXPORT_OK %EXPORT_TAGS $DEBUG);
 
 $DEBUG = 0;
-$VERSION = '0.48';
+our $VERSION = '0.52';
 
 use base qw(Class::Accessor);
 XML::Sig->mk_accessors(qw(key));
@@ -21,6 +23,7 @@ use base qw/Exporter/;
 
 
 use Digest::SHA qw(sha1 sha224 sha256 sha384 sha512 hmac_sha1 hmac_sha256 hmac_sha384 hmac_sha512);
+use Crypt::Digest::RIPEMD160 qw/ripemd160/;
 use XML::LibXML;
 use MIME::Base64;
 use Carp;
@@ -67,20 +70,20 @@ sub new {
         $self->_load_cert_text( $params->{ 'cert_text' } );
     }
 
-    if ( exists $params->{ sig_hash } && grep { $_ eq $params->{ sig_hash } } ('sha224', 'sha256', 'sha384', 'sha512'))
+    if ( exists $params->{ sig_hash } && grep { $_ eq $params->{ sig_hash } } ('sha224', 'sha256', 'sha384', 'sha512', 'ripemd160'))
     {
         $self->{ sig_hash } = $params->{ sig_hash };
     }
     else {
-        $self->{ sig_hash } = 'sha1';
+        $self->{ sig_hash } = 'sha256';
     }
 
-    if ( exists $params->{ digest_hash } && grep { $_ eq $params->{ digest_hash } } ('sha1', 'sha224', 'sha256', 'sha384',, 'sha512'))
+    if ( exists $params->{ digest_hash } && grep { $_ eq $params->{ digest_hash } } ('sha1', 'sha224', 'sha256', 'sha384','sha512', 'ripemd160'))
     {
         $self->{ digest_hash } = $params->{ digest_hash };
     }
     else {
-        $self->{ digest_hash } = 'sha1';
+        $self->{ digest_hash } = 'sha256';
     }
 
     if (defined $self->{ key_type } && $self->{ key_type } eq 'dsa') {
@@ -88,7 +91,7 @@ sub new {
             $self->{ sig_hash } = $params->{ sig_hash };
         }
         else {
-            $self->{ sig_hash } = 'sha1';
+            $self->{ sig_hash } = 'sha256';
         }
     }
 
@@ -158,12 +161,15 @@ sub sign {
         if(my $ref = Digest::SHA->can($self->{ digest_hash })) {
             $self->{digest_method} = $ref;
         }
+        elsif ( $ref = Crypt::Digest::RIPEMD160->can($self->{ digest_hash }))  {
+            $self->{digest_method} = $ref;
+        }
         else {
             die("Can't handle $self->{ digest_hash }");
         }
 
         # Calculate the digest of the XML being signed
-        my $bin_digest    = $self->{digest_method}->( $xml_canon );
+        my $bin_digest    = $self->{digest_method}->( Encode::encode_utf8( $xml_canon ));
         my $digest        = encode_base64( $bin_digest, '' );
         print ("   Digest: $digest\n") if $DEBUG;
 
@@ -195,63 +201,14 @@ sub sign {
         # Calculate the signature of the Canonical Form of SignedInfo
         my $signature;
         if ($self->{key_type} eq 'dsa') {
-            print ("    Signing SignedInfo using DSA key type\n") if $DEBUG;
-            if(my $ref = Digest::SHA->can($self->{ sig_hash })) {
-                $self->{sig_method} = $ref;
-            }
-            else {
-                die("Can't handle $self->{ sig_hash }");
-            }
-
-            # DSA 1024-bit only permits the signing of 20 bytes or less, hence the sha1
-            # DSA 2048-bit only permits the signing sha256
-            my $bin_signature = $self->{key_obj}->do_sign( $self->{ sig_method }($signed_info_canon) );
-
-            # https://www.w3.org/TR/2002/REC-xmldsig-core-20020212/#sec-SignatureAlg
-            # The output of the DSA algorithm consists of a pair of integers
-            # The signature value consists of the base64 encoding of the
-            # concatenation of r and s in that order ($r . $s)
-            my $r = $bin_signature->get_r;
-            my $s = $bin_signature->get_s;
-
-            my $sig_size = ($self->{key_obj}->get_sig_size - 8) * 8;
-            my $rs = _zero_fill_buffer($sig_size);
-            _concat_dsa_sig_r_s(\$rs, $r, $s, $sig_size);
-
-            $signature        = encode_base64( $rs, "\n" );
+            $signature = encode_base64( $self->_calc_dsa_signature( $signed_info_canon ), "\n" );
         } elsif ($self->{key_type} eq 'ecdsa') {
-            print ("    Signing SignedInfo using ECDSA key type\n") if $DEBUG;
-
-            my $bin_signature = $self->{key_obj}->sign_message_rfc7518(
-                $signed_info_canon, uc($self->{sig_hash})
-            );
-            # The output of the ECDSA algorithm consists of a pair of integers
-            # The signature value consists of the base64 encoding of the
-            # concatenation of r and s in that order ($r . $s).  In this
-            # case sign_message_rfc7518 produces that
-
-            $signature        = encode_base64( $bin_signature, "\n" );
+            $signature = encode_base64( $self->_calc_ecdsa_signature( $signed_info_canon ), "\n" );
         } elsif ($self->{key_type} eq 'rsa') {
-            print ("    Signing SignedInfo using RSA key type\n") if $DEBUG;
-            my $sig_hash = 'use_' . $self->{ sig_hash } . '_hash';
-            $self->{key_obj}->$sig_hash;
-            my $bin_signature = $self->{key_obj}->sign( $signed_info_canon );
-            $signature        = encode_base64( $bin_signature, "\n" );
+            $signature = encode_base64( $self->_calc_rsa_signature( $signed_info_canon ), "\n" );
         } else {
             if ( defined $self->{ hmac_key } ) {
-                print ("    Signing SignedInfo using hmac-", $self->{ sig_hash }, "\n") if $DEBUG;
-                if (my $ref = Digest::SHA->can('hmac_' . $self->{ sig_hash })) {
-                    $self->{sig_method} = $ref;
-                }
-                else {
-                    die("Can't handle $self->{ sig_hash }");
-                }
-
-                my $bin_signature = $self->{sig_method} (
-                                        $signed_info_canon,
-                                        decode_base64( $self->{ hmac_key } )
-                                    );
-                $signature        = encode_base64( $bin_signature, "\n" );
+                $signature = encode_base64( $self->_calc_hmac_signature( $signed_info_canon ), "\n" );
             } else {
                 die "No Signature signing method provided";
             }
@@ -353,6 +310,9 @@ sub verify {
         if(my $ref = Digest::SHA->can($signature_method)) {
             $self->{sig_method} = $ref;
         }
+        elsif ( $ref = Crypt::Digest::RIPEMD160->can( $signature_method ))  {
+            $self->{sig_method} = $ref;
+        }
         else {
             die("Can't handle $signature_method");
         }
@@ -360,6 +320,9 @@ sub verify {
         if(my $ref = Digest::SHA->can($digest_method)) {
             $self->{digest_method} = $ref;
         }
+        elsif ( $ref = Crypt::Digest::RIPEMD160->can( $digest_method ))  {
+            $self->{digest_method} = $ref;
+        }
         else {
             die("Can't handle $digest_method");
         }
@@ -433,8 +396,10 @@ sub verify {
         # signatures to be validated if they exist
         $signed_xml->addChild( $signature_node );
 
+        print ( "    Canonical XML:  " . $canonical ."\n") if $DEBUG;
+
         # Obtain the DigestValue of the Canonical XML
-        my $digest = $self->{digest_method}->($canonical);
+        my $digest = $self->{digest_method}->(Encode::encode_utf8($canonical));
 
         print ( "    Reference Digest:  " . _trim($refdigest) ."\n") if $DEBUG;
 
@@ -913,7 +878,7 @@ sub _verify_ecdsa {
     or confess "Crypt::PK::ECC 0.036+ needs to be installed so
              that we can handle ECDSA signatures";
     # Generate Public Key from XML
-    my $oid = _trim($self->{parser}->findvalue('//dsig:NamedCurve/@URN', $context));
+    my $oid = _trim($self->{parser}->findvalue('.//dsig:NamedCurve/@URN', $context));
 
     use URI ();
     my $u1 = URI->new($oid);
@@ -934,8 +899,8 @@ sub _verify_ecdsa {
         '1.3.36.3.3.2.8.1.1.13' => 'brainpoolP512r1',
     );
 
-    my $x = $self->{parser}->findvalue('//dsig:PublicKey/dsig:X/@Value', $context);
-    my $y = $self->{parser}->findvalue('//dsig:PublicKey/dsig:Y/@Value', $context);
+    my $x = $self->{parser}->findvalue('.//dsig:PublicKey/dsig:X/@Value', $context);
+    my $y = $self->{parser}->findvalue('.//dsig:PublicKey/dsig:Y/@Value', $context);
 
     my $ecdsa_pub = Crypt::PK::ECC->new();
 
@@ -975,22 +940,29 @@ sub _verify_hmac {
 
     # Decode signature and verify
     my $bin_signature = decode_base64($sig);
-
+    use Crypt::Mac::HMAC qw( hmac );
     if ( defined $self->{ hmac_key } ) {
         print ("    Verifying SignedInfo using hmac-", $self->{ sig_hash }, "\n") if $DEBUG;
         if ( my $ref = Digest::SHA->can('hmac_' . $self->{ sig_hash }) ) {
-            $self->{sig_method} = $ref;
+            if ($bin_signature eq $self->_calc_hmac_signature( $canonical )) {
+                return 1;
+            }
+            else {
+                return 0;
+            }
+        }
+        elsif ( $ref = Crypt::Digest::RIPEMD160->can($self->{ sig_hash })) {
+            if ($bin_signature eq $self->_calc_hmac_signature( $canonical )) {
+                return 1;
+            }
+            else {
+                return 0;
+            }
         }
         else {
             die("Can't handle $self->{ sig_hash }");
         }
 
-        if ($bin_signature eq $self->{sig_method}( $canonical, decode_base64( $self->{ hmac_key } ))) {
-            return 1;
-        }
-        else {
-            return 0;
-        }
     } else {
         return 0;
     }
@@ -1433,7 +1405,12 @@ sub _signedinfo_xml {
         $algorithm = "http://www.w3.org/2000/09/xmldsig#$self->{key_type}-$self->{ sig_hash }";
     }
     elsif ( $self->{key_type} eq 'ecdsa' ) {
-        $algorithm = "http://www.w3.org/2001/04/xmldsig-more#$self->{key_type}-$self->{ sig_hash }";
+        if ( $self->{ sig_hash } eq 'ripemd160' || $self->{ sig_hash } eq  'whirlpool' ) {
+            $algorithm = "http://www.w3.org/2007/05/xmldsig-more#$self->{key_type}-$self->{ sig_hash }";
+        }
+        else {
+            $algorithm = "http://www.w3.org/2001/04/xmldsig-more#$self->{key_type}-$self->{ sig_hash }";
+        }
     }
     elsif ( $self->{ key_type } eq 'dsa' && $self->{ sig_hash } eq 'sha256') {
         $algorithm = "http://www.w3.org/2009/xmldsig11#$self->{key_type}-$self->{ sig_hash }";
@@ -1444,7 +1421,7 @@ sub _signedinfo_xml {
 
     #return qq{<dsig:SignedInfo xmlns:dsig="http://www.w3.org/2000/09/xmldsig#">
     return qq{<dsig:SignedInfo xmlns:dsig="http://www.w3.org/2000/09/xmldsig#" xmlns:xenc="http://www.w3.org/2001/04/xmlenc#">
-                <dsig:CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments" />
+                <dsig:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#" />
                 <dsig:SignatureMethod Algorithm="$algorithm" />
                 $digest_xml
             </dsig:SignedInfo>};
@@ -1543,6 +1520,133 @@ sub _canonicalize_xml {
     }
     return $xml;
 }
+
+##
+## _calc_dsa_signature($signed_info_canon)
+##
+## Arguments:
+##    $canonical:     string Canonical XML
+##
+## Returns: string  Signature
+##
+## Calculates signature based on the method and hash
+##
+sub _calc_dsa_signature {
+    my $self                = shift;
+    my $signed_info_canon   = shift;
+
+    print ("    Signing SignedInfo using DSA key type\n") if $DEBUG;
+    if(my $ref = Digest::SHA->can($self->{ sig_hash })) {
+        $self->{sig_method} = $ref;
+    }
+    elsif ( $ref = Crypt::Digest::RIPEMD160->can($self->{ sig_hash }))  {
+        $self->{sig_method} = $ref;
+    }
+    else {
+        die("Can't handle $self->{ sig_hash }");
+    }
+
+    # DSA 1024-bit only permits the signing of 20 bytes or less, hence the sha1
+    # DSA 2048-bit only permits the signing sha256
+    my $bin_signature = $self->{key_obj}->do_sign( $self->{ sig_method }($signed_info_canon) );
+
+    # https://www.w3.org/TR/2002/REC-xmldsig-core-20020212/#sec-SignatureAlg
+    # The output of the DSA algorithm consists of a pair of integers
+    # The signature value consists of the base64 encoding of the
+    # concatenation of r and s in that order ($r . $s)
+    my $r = $bin_signature->get_r;
+    my $s = $bin_signature->get_s;
+
+    my $sig_size = ($self->{key_obj}->get_sig_size - 8) * 8;
+    my $rs = _zero_fill_buffer($sig_size);
+    _concat_dsa_sig_r_s(\$rs, $r, $s, $sig_size);
+
+    return $rs;
+
+}
+
+##
+## _calc_ecdsa_signature($signed_info_canon)
+##
+## Arguments:
+##    $canonical:     string Canonical XML
+##
+## Returns: string  Signature
+##
+## Calculates signature based on the method and hash
+##
+sub _calc_ecdsa_signature {
+    my $self                = shift;
+    my $signed_info_canon   = shift;
+
+    print ("    Signing SignedInfo using ECDSA key type\n") if $DEBUG;
+
+    my $bin_signature = $self->{key_obj}->sign_message_rfc7518(
+        $signed_info_canon, uc($self->{sig_hash})
+    );
+    # The output of the ECDSA algorithm consists of a pair of integers
+    # The signature value consists of the base64 encoding of the
+    # concatenation of r and s in that order ($r . $s).  In this
+    # case sign_message_rfc7518 produces that
+    return $bin_signature;
+}
+
+##
+## _calc_rsa_signature($signed_info_canon)
+##
+## Arguments:
+##    $canonical:     string Canonical XML
+##
+## Returns: string  Signature
+##
+## Calculates signature based on the method and hash
+##
+sub _calc_rsa_signature {
+    my $self                = shift;
+    my $signed_info_canon   = shift;
+
+    print ("    Signing SignedInfo using RSA key type\n") if $DEBUG;
+    my $sig_hash = 'use_' . $self->{ sig_hash } . '_hash';
+    $self->{key_obj}->$sig_hash;
+    my $bin_signature = $self->{key_obj}->sign( $signed_info_canon );
+
+    return $bin_signature;
+}
+
+##
+## _calc_hmac_signature($signed_info_canon)
+##
+## Arguments:
+##    $signed_info_canon:     string Canonical XML
+##
+## Returns: string  Signature
+##
+## Calculates signature based on the method and hash
+##
+sub _calc_hmac_signature {
+    my $self                = shift;
+    my $signed_info_canon   = shift;
+
+    use Crypt::Mac::HMAC qw( hmac );
+    my $bin_signature;
+    print ("    Signing SignedInfo using hmac-", $self->{ sig_hash }, "\n") if $DEBUG;
+    if (my $ref = Digest::SHA->can('hmac_' . $self->{ sig_hash })) {
+        $self->{sig_method} = $ref;
+        $bin_signature = $self->{sig_method} (
+                            $signed_info_canon,
+                            decode_base64( $self->{ hmac_key } )
+                        );
+    }
+    elsif ( $ref = Crypt::Digest::RIPEMD160->can($self->{ sig_hash }))  {
+        $self->{sig_method} = $ref;
+        $bin_signature = hmac('RIPEMD160', decode_base64( $self->{ hmac_key } ), $signed_info_canon );
+    }
+    else {
+        die("Can't handle $self->{ sig_hash }");
+    }
+
+    return $bin_signature;
+}
 1;
 
 =pod
@@ -1551,11 +1655,11 @@ sub _canonicalize_xml {
 
 =head1 NAME
 
-XML::Sig
+XML::Sig - XML::Sig - A toolkit to help sign and verify XML Digital Signatures
 
 =head1 VERSION
 
-version 0.48
+version 0.52
 
 =head1 SYNOPSIS
 
@@ -1694,7 +1798,7 @@ anything with DSA anyway).
 Passing digest_hash to new allows you to specify the DigestMethod
 hashing algorithm used when calculating the hash of the XML being
 signed.  Supported hashes can be specified sha1, sha224, sha256,
-sha384, and sha512
+sha384, sha512, ripemd160
 
 =item B<hmac_key>
 
